{ Main view, where most of the application logic takes place.

  Feel free to use this code as a starting point for your own projects.
  This template code is in public domain, unlike most other CGE code which
  is covered by BSD or LGPL (see https://castle-engine.io/license). }
unit GameViewMain;

{$MODE DELPHI}

interface

uses Classes,
  CastleVectors, CastleComponentSerialize,
  CastleUIControls, CastleControls, CastleKeysMouse, CastleScene, castletransform, castleviewport, castlecolors,
  x3dnodes, castleimages, CastleShapes, castlecameras;

type
  { Main view, where most of the application logic takes place. }
  TViewMain = class(TCastleView)
  published
    { Components designed using CGE editor.
      These fields will be automatically initialized at Start. }
    LabelFps: TCastleLabel;
    Viewport1: TCastleViewport;
    ExamineNavigation1: TCastleExamineNavigation;    
  public
    constructor Create(AOwner: TComponent); override;
    procedure Start; override;
    procedure Update(const SecondsPassed: Single; var HandleInput: Boolean); override;
    function Press(const Event: TInputPressRelease): Boolean; override;
  private
    FScene: TCastleTransform;
    function GetSceneUnderMouseWithPixelTest(): TCastleTransform;
    procedure SetColor(AColor: TCastleColor);
  end;

var
  ViewMain: TViewMain;

implementation

uses SysUtils, generics.collections, CastleStringUtils;

{ TViewMain ----------------------------------------------------------------- }

function IIF(ACond: Boolean; IfTrue, IfFalse: string): string;
begin
  if ACond then
    Result := IfTrue  
  else
    Result := IfFalse;
end;

constructor TViewMain.Create(AOwner: TComponent);
begin
  inherited;
  DesignUrl := 'castle-data:/gameviewmain.castle-user-interface';
end;

procedure TViewMain.Start;
begin
  inherited;
  ExamineNavigation1.TurnTable := True;
end;

procedure TViewMain.SetColor(AColor: TCastleColor);
begin
  if FScene is TCastleScene then
    TCastleScene(FScene).RenderOptions.WireframeColor := AColor.RGB
  else if FScene is TCastleImageTransform then
    TCastleImageTransform(FScene).Color := AColor;
end;

procedure TViewMain.Update(const SecondsPassed: Single; var HandleInput: Boolean);
var LScene: TCastleTransform;
begin
  inherited;
  { This virtual method is executed every frame (many times per second). }
  Assert(LabelFps <> nil, 'If you remove LabelFps from the design, remember to remove also the assignment "LabelFps.Caption := ..." from code');
  LabelFps.Caption := 'FPS: ' + Container.Fps.ToString;


  ///////////////////////////// the place to change the implementation /////////////////////////////
  LScene := GetSceneUnderMouseWithPixelTest();

  
  if Assigned(LScene) then 
  begin
    if Assigned(FScene) and (FScene <> LScene) then 
      SetColor(Black);
    FScene := LScene;
    SetColor(Red);
  end
  else
  begin
    if Assigned(FScene) then
      SetColor(Black); 
    FScene := nil;
  end;
end;

function TViewMain.GetSceneUnderMouseWithPixelTest(): TCastleTransform;
var
  LRay: TRayCollision;
  LRayNode: TRayCollisionNode;
  LOrigin: TVector3;
  LDirection: TVector3;
  LTexCoords: TVector2;
  LHitLocalCoords: TVector3;
  LCounter, LChildIndex: Integer;
  function IsPixelTransparent(): Boolean;
  begin
    Result := False;

    // LRayNode.Point is in coordinates of the LRayNode.Item.Parent
    LHitLocalCoords := LRayNode.Item.InverseTransform.MultPoint(LRayNode.Point);
    // ITexCoord2D expects coordinates in the local space of this TCastleTransform
    LTexCoords := LRayNode.Triangle^.ITexCoord2D(LHitLocalCoords);
    LabelFps.Caption := Format('%s| P %s| TC %s', [LabelFps.Caption, LRayNode.Point.ToString, LTexCoords.ToString]);

    Result := LRayNode.Triangle^.State.MainTexture.Color(LTexCoords)[3] < 0.2; // Consider pixel transparent if alpha is less than 0.2(*255=51)  
  end;
  function PtrToStr(APointer: Pointer): string;
  begin
    Result := Copy(PointerToStr(APointer), 12, 8);
  end;
begin
  Result := nil;
  if not Assigned(Viewport1.MouseRayHit) then
    Exit;
  Viewport1.PositionToRay(Container.MousePosition, True, LOrigin, LDirection);
  LCounter := 0;
  repeat
    Inc(LCounter);
    LRay := Viewport1.Items.WorldRay(LOrigin, LDirection);
    if not Assigned(LRay) or not LRay.Info(LRayNode) then
      Exit;
    LabelFps.Caption := Format('%s| MR %s O%s D %s Ray %s RN %s T %s C %d',[LabelFps.Caption, Viewport1.MouseRayHit[0].Point.XY.ToString,
      '-'{LOrigin.ToString}, LDirection.ToString, PtrToStr(LRay), PtrToStr(@LRayNode), PtrToStr(LRayNode.Triangle), LRay.Count]);
    
    for LChildIndex := 0 to LRay.Count - 1 do
    begin
      LRayNode := LRay.Items[LChildIndex];
      if not Assigned(LRayNode.Triangle) then       
        Continue;
      LabelFps.Caption := Format('%s| TxC', [LabelFps.Caption]);
      if not Assigned(LRayNode.Triangle^.State.MainTexture) then        
        Continue;
      LabelFps.Caption := Format('%s| PC', [LabelFps.Caption]);
      if not IsPixelTransparent() then
        Exit(LRayNode.Item);
    end; 

    LOrigin := LRayNode.Point + LDirection.AdjustToLength(SingleEpsilon);
    Result := LRay.Transform;

  until LCounter > 100;
end;


function TViewMain.Press(const Event: TInputPressRelease): Boolean;
begin
  Result := inherited;
  if Result then Exit; // allow the ancestor to handle keys

  { This virtual method is executed when user presses
    a key, a mouse button, or touches a touch-screen.

    Note that each UI control has also events like OnPress and OnClick.
    These events can be used to handle the "press", if it should do something
    specific when used in that UI control.
    The TViewMain.Press method should be used to handle keys
    not handled in children controls.
  }

  // Use this to handle keys:
  {
  if Event.IsKey(keyXxx) then
  begin
    // DoSomething;
    Exit(true); // key was handled
  end;
  }
end;

end.
